;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "cargo-mode" "0.0.10"
  "Cargo Major Mode. Cargo is the Rust package manager."
  '((emacs "25.1"))
  :url "https://github.com/ayrat555/cargo-mode"
  :commit "33528954218f8957a26f3fef506c3537823d569d"
  :revdesc "33528954218f"
  :keywords '("tools")
  :authors '(("Ayrat Badykov" . "ayratin555@gmail.com"))
  :maintainers '(("Ayrat Badykov" . "ayratin555@gmail.com")))
